//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"context"
	"errors"
	"net/http"
	"strings"
	"sync"
	"time"
)

const (
	defaultAge     = 24 * time.Hour
	tokenSeperator = ':'
)

type SessionState int

const (
	SessionStateNew SessionState = iota
	SessionStateLoginInProgress
	SessionStateLoggedIn
	SessionStateLoginFailed
)

func (s SessionState) String() string {
	switch s {
	case SessionStateNew:
		return "new"
	case SessionStateLoginInProgress:
		return "login-in-progress"
	case SessionStateLoggedIn:
		return "logged-in"
	case SessionStateLoginFailed:
		return "login-failed"
	}
	return "unknown"
}

func (s SessionState) MarshalText() (data []byte, err error) {
	data = []byte(s.String())
	return
}

type Session struct {
	secret   string
	ID       string       `json:"id"`
	State    SessionState `json:"state"`
	Expires  time.Time    `json:"expires"`
	Username string       `json:"username"`
	ReadOnly bool         `json:"readonly"`
	AllShows bool         `json:"all-shows"`
	Shows    []string     `json:"shows"`

	subscribe chan struct{}
	oidc      *OIDCSession
}

func NewSession() (s *Session, err error) {
	s = &Session{}
	if s.ID, err = generateRandomString(16); err != nil {
		return
	}
	if s.secret, err = generateRandomString(32); err != nil {
		return
	}
	s.State = SessionStateNew
	return
}

var (
	anonAllowNone  = &Session{Username: "anonymous", ReadOnly: false, AllShows: false, Shows: []string{}}
	anonAllowAll   = &Session{Username: "anonymous", ReadOnly: false, AllShows: true, Shows: []string{}}
	anonAllowAllRO = &Session{Username: "anonymous", ReadOnly: true, AllShows: true, Shows: []string{}}
)

func (s *Session) Expired() bool {
	return s.Expires.Before(time.Now())
}

func (s *Session) getToken() string {
	return s.ID + string(tokenSeperator) + s.secret
}

func parseBearerAuthHeader(authHeader string) (string, string, bool) {
	const prefix = "Bearer "

	if len(authHeader) < len(prefix) || !strings.EqualFold(authHeader[:len(prefix)], prefix) {
		return "", "", false
	}
	token := authHeader[len(prefix):]
	if token == "" {
		return "", "", false
	}
	sepIdx := strings.IndexByte(token, tokenSeperator)
	if sepIdx < 0 {
		return "", "", false
	}
	return token[:sepIdx], token[sepIdx+1:], true
}

func getSessionFromBearerToken(r *http.Request) *Session {
	authHeader := r.Header.Get("Authorization")
	if authHeader == "" {
		return nil
	}

	sID, secret, ok := parseBearerAuthHeader(authHeader)
	if !ok {
		return nil
	}

	s := auth.sessions.get(sID)
	if s == nil || s.secret != secret {
		return nil
	}
	if s.Expired() {
		auth.sessions.remove(sID)
		return nil
	}

	return s
}

func attachSessionToRequest(r *http.Request, s *Session) *http.Request {
	ctx := context.WithValue(r.Context(), sessionContextKey, s)
	return r.WithContext(ctx)
}

func SessionFromRequest(r *http.Request) (*Session, bool) {
	s, ok := r.Context().Value(sessionContextKey).(*Session)
	return s, ok
}

type SessionManager struct {
	mutex    sync.RWMutex
	maxAge   time.Duration
	sessions map[string]*Session
}

func NewSessionManager(c SessionsConfig) (sm *SessionManager, err error) {
	sm = &SessionManager{maxAge: defaultAge}
	if c.MaxAge > 0 {
		sm.maxAge = c.MaxAge
	}
	sm.sessions = make(map[string]*Session)
	go sm.runMaintenance()
	return
}

func (sm *SessionManager) runMaintenance() {
	t := time.NewTicker(time.Minute * 10)
	for {
		<-t.C
		sm.cleanup()
	}
}

func (sm *SessionManager) insert(s *Session) (err error) {
	now := time.Now()
	eMax := now.Add(sm.maxAge)
	if s.Expires.Before(now) || s.Expires.After(eMax) {
		s.Expires = eMax
	}

	sm.mutex.Lock()
	defer sm.mutex.Unlock()
	sm.sessions[s.ID] = s
	return
}

func (sm *SessionManager) get(id string) *Session {
	sm.mutex.RLock()
	defer sm.mutex.RUnlock()

	s, ok := sm.sessions[id]
	if !ok {
		return nil
	}
	return s
}

func (sm *SessionManager) getAndSubscribe(id string) (*Session, <-chan struct{}) {
	sm.mutex.RLock()
	defer sm.mutex.RUnlock()

	s, ok := sm.sessions[id]
	if !ok {
		return nil, nil
	}
	if s.subscribe == nil {
		s.subscribe = make(chan struct{})
	}
	return s, s.subscribe
}

func (sm *SessionManager) update(id string, s *Session) error {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	old, ok := sm.sessions[id]
	if !ok {
		return errors.New("session not found.")
	}
	if s.Expires.IsZero() {
		s.Expires = old.Expires
	}
	s.secret = old.secret
	s.ID = old.ID
	if old.subscribe != nil {
		close(old.subscribe)
	}
	sm.sessions[id] = s
	return nil
}

// TODO: this is racy!!
func (sm *SessionManager) updateState(id string, st SessionState) error {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	s, ok := sm.sessions[id]
	if !ok {
		return errors.New("session not found.")
	}
	if s.subscribe != nil {
		close(s.subscribe)
		s.subscribe = nil
	}
	s.State = st
	return nil
}

func (sm *SessionManager) remove(id string) {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	delete(sm.sessions, id)
}

func (sm *SessionManager) cleanup() {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	for id, s := range sm.sessions {
		if s.Expired() {
			delete(sm.sessions, id)
		}
	}
}
